/* scope.c   by Michael Thorpe   2006-03-12 */

#include <stdlib.h>
#include <string.h>
#include "lisp.h"

typedef struct entry entry;

struct entry {
	entry *next;
	obj *name;
	obj *value;
};

#ifdef DEBUG
unsigned long total_scopes=0;
#endif

static scope topscope={1,0,0};
scope *globalscope=0;

static obj *already_defined_error,*undefined_error;

/* real_lookup doesn't decref/incref anything */
static entry *real_lookup(scope *scope,obj *name) {
	entry *e;

	if(!scope)
		scope=globalscope;
	while(scope) {
		for(e=scope->firstentry;e;e=e->next)
			if(e->name==name)
				return(e);
		scope=scope->parent;
	}
	return(0);
}

obj *define(obj *name,obj *value) {
	entry *e;

	e=real_lookup(globalscope,name);
	if(e) {
		decref(value);
		return(throw(incref(already_defined_error),name));
	}
	if(real_define(globalscope,name,value)) {
		decref(value);
		decref(name);
		return(0);
	}
	decref(value);
	return(name);
}

void drop_scope(scope *s) {
	scope *t;
	entry *e,*f;

#ifdef DEBUG
	total_scopes--;
#endif
	e=s->firstentry;
	while(e) {
		f=e;
		e=e->next;
		decref(f->name);
		decref(f->value);
		free(f);
	}
	t=s->parent;
#ifdef DEBUG
	memset(s,'s',sizeof(scope));
#else
	free(s);
#endif
	decscoperef(t);
}

static obj *let(scope *scope,obj *var,obj *val,obj *block) {
	/* We can short-circuit the scope creation if its refcount is 1: */
	if(1 != scope->refcount) {
		scope=newscope(scope);
		if(!scope) {
			decref(var);
			decref(val);
			decref(block);
			return(0);
		}
	}
	if(real_define(scope,var,val)) {
		decref(var);
		decref(val);
		decscoperef(scope);
		decref(block);
		return(0);
	}
	decref(var);
	decref(val);
	return(evaluate(scope,block));
}

obj *lookup(scope *scope,obj *name) {
	entry *e;
	obj *val;

	e=real_lookup(scope,name);
	if(!e) {
		decscoperef(scope);
		return(throw(incref(undefined_error),name));
	}
	decref(name);
	val=incref(e->value);
	decscoperef(scope);
	return(val);
}

scope *newscope(scope *parent) {
	scope *new;

	if(!parent)
		parent=incscoperef(globalscope);
	new=(scope *)malloc(sizeof(scope));
	if(!new) {
		decscoperef(parent);
		throwoom();
		return(0);
	}
#ifdef DEBUG
	total_scopes++;
#endif
	new->refcount=1;
	new->parent=parent;
	new->firstentry=0;
	return(new);
}

int real_define(scope *s,obj *name,obj *value) {
	entry *e;

	e=(entry *)malloc(sizeof(entry));
	if(!e) {
		throwoom();
		return(-1);
	}
	e->next=s->firstentry;
	e->name=incref(name);
	e->value=incref(value);
	s->firstentry=e;
	return(0);
}

obj *set(scope *s,obj *name,obj *value) {
	entry *e;

	e=real_lookup(s,name);
	decscoperef(s);
	if(!e) {
		decref(value);
		return(throw(incref(undefined_error),name));
	}
	decref(e->value);
	e->value=value;
	return(name);
}

int store_builtin(unsigned int numargs,unsigned int wantscope,const char *name,obj *(*rawfunc)()) {
	obj *key,*func;
	int i;

	key=newword(name,strlen(name));
	if(!key)
		return(-1);
	func=newobj(&rawfunc_objtype);
	if(!func) {
		decref(key);
		return(-1);
	}
	func->value.rawfunc.numargs=numargs;
	func->value.rawfunc.wantscope=wantscope;
	func->value.rawfunc.rawfuncptr.ns0=rawfunc;
	i=real_define(globalscope,key,func);
	decref(key);
	decref(func);
	return(i);
}

int store_object(const char *name,obj *o) {
	obj *key;
	int i;

	key=newword(name,strlen(name));
	if(!key)
		return(-1);
	i=real_define(globalscope,key,o);
	decref(key);
	decref(o);
	return(i);
}

int scope_init() {
	if(!(globalscope=newscope(incscoperef(&topscope)))
	|| !(already_defined_error=newword("already-defined",15))
	|| !(undefined_error=newword("undefined",9))
	|| store_builtin_ns2("#define",define)
	|| store_builtin_s3("#let",let)
	|| store_builtin_s2("#set",set))
		return(1);
	INC_KNOWN_OBJECTS(2);
	return(0);
}

